using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Threading.Tasks;
using WingInterfaceLibrary.DTO.User;
using WingInterfaceLibrary.Enum;
using WingInterfaceLibrary.Internal.Interface;
using WingInterfaceLibrary.Internal.Request;
using WingServerCommon.Config;
using WingServerCommon.Config.Parameters;
using WingServerCommon.Log;

namespace WingUserModule.Common
{
    public class ScanLoginManager
    {


        public static IList<string> ExpireOrErrorScanCodes = new List<string>();

        private IMasterInteractionCenterService _masterInteractionCenterService;

        private ConcurrentDictionary<string, UserScanInfo> _userInfos = new ConcurrentDictionary<string, UserScanInfo>();

        public ScanLoginManager(IMasterInteractionCenterService masterInteractionCenterService)
        {
            _masterInteractionCenterService = masterInteractionCenterService;
            ExpireScanUserWatching();
        }

        public async Task<bool> AddOrUpdate(UserScanInfo userScanInfo, int syncType = 0)
        {

            _userInfos.AddOrUpdate(userScanInfo.ScanCode, userScanInfo, (k, v) =>
            {
                return userScanInfo;
            });
            //同步其他服务器
            if (syncType == 0)
            {
                await SyncToMasterAsync(userScanInfo);
            }
            return true;
        }

        /// <summary>
        /// 删除
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public bool Remove(string key)
        {
            return _userInfos.TryRemove(key, out _);
        }

        /// <summary>
        /// 获取
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public UserScanInfo Get(string key)
        {
            if (_userInfos.TryGetValue(key, out UserScanInfo obj))
            {
                return obj;
            }
            return null;
        }

       

        private async Task ExpireScanUserWatching()
        {
            Logger.WriteLineInfo($"ExpireScanUserWatching start");
            Task.Run(async () =>
            {
                while (true)
                {
                    await Task.Delay(1000 * 10);
                    foreach (var scanCode in _userInfos.Keys)
                    {
                        var userInfo = _userInfos[scanCode];
                        try
                        {
                            if (userInfo.ExpireTime < DateTime.UtcNow.AddSeconds(-10))
                            {
                                Remove(userInfo.ScanCode);
                            }
                        }
                        catch (Exception ex)
                        {
                            Logger.WriteLineWarn($"cExpireScanUserWatching {userInfo.ScanCode} err, {ex}");
                        }
                        finally
                        {
                            if (userInfo.ExpireTime < DateTime.UtcNow.AddSeconds(-10))
                            {
                                Remove(userInfo.ScanCode);
                            }
                        }
                    }
                }
            });
            Logger.WriteLineInfo($"ExpireScanUserWatching finished");
        }

        /// <summary>
        /// 同步内存给主服务器
        /// </summary>
        /// <param name="userScanInfo"></param>
        /// <returns></returns>
        private async Task SyncToMasterAsync(UserScanInfo userScanInfo)
        {
            try
            {
                if (ConfigurationManager.IsDistributed)
                {
                    var syncRequest = new SyncReceiveServiceDataRequest
                    {
                        SyncService = SyncServiceEnum.ScanLogin,
                        SourceUrl = EnvironmentConfigManager.GetParammeter<StringParameter>("Gateway", "Domains").Value,
                        ServerID = EnvironmentConfigManager.GetParammeter<StringParameter>("General", "ServerID").Value,
                        ServiceDataJson = Newtonsoft.Json.JsonConvert.SerializeObject(userScanInfo),
                    };
                    await _masterInteractionCenterService.SyncReceiveSlaveServiceDataAsync(syncRequest);
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineWarn($"ScanLoginManager SyncToMasterAsync err, userScanInfo:{Newtonsoft.Json.JsonConvert.SerializeObject(userScanInfo)}, err: {ex}");
            }
        }
    }

}