﻿using System.Threading.Tasks;
using System.Collections.Generic;
using WingInterfaceLibrary.Interface;
using JsonRpcLite.Services;
using WingInterfaceLibrary.Request.User;
using WingInterfaceLibrary.DTO.Role;
using WingInterfaceLibrary.Request;
using WingInterfaceLibrary.Enum;
using WingServerCommon.Service;
using System.Linq;

namespace WingUserModule.Service
{
    public partial class UserService : JsonRpcService, IRoleService
    {
        /// <summary>
        /// 获取角色信息
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<RoleDTO> GetRoleByCodeAsync(GetRoleByCodeRequest request)
        {
            if (string.IsNullOrWhiteSpace(request.RoleCode))
            {
                ThrowCustomerException(CustomerRpcCode.RoleCodesIsEmpty, "RoleCodesIsEmpty");
            }
            var role = await _roleDBService.FindRoleByCodeAsync(request.RoleCode);
            if (role == null)
            {
                ThrowCustomerException(CustomerRpcCode.RoleNoExist, "RoleNoExist");
            }
            ChangeRoleByLanguageConfig(role, request.Language);
            return role;
        }

        /// <summary>
        /// 查找默认角色
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<List<RoleDTO>> FindDefaultRolesAsync(FindDefaultRolesRequest request)
        {
            var roles = await _roleDBService.FindRolesByTypeAsync(RoleShowTypeEnum.NotShow, request.OrganizationCode);
            if (roles != null && roles.Count > 0)
            {
                foreach (var role in roles)
                {
                    ChangeRoleByLanguageConfig(role, request.Language);
                }
            }
            return roles;
        }

        /// <summary>
        /// 查找授权角色
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<List<RoleDTO>> FindAuthenticationRolesAsync(FindAuthenticationRolesRequest request)
        {
            var roles = await _roleDBService.FindRolesByTypeAsync(RoleShowTypeEnum.ISShow, "");
            if (roles != null && roles.Count > 0)
            {
                foreach (var role in roles)
                {
                    ChangeRoleByLanguageConfig(role, request.Language);
                }
            }
            return roles;
        }

        /// <summary>
        /// 查找全部角色
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<List<RoleDTO>> FindAllRolesAsync(FindAuthenticationRolesRequest request) 
        {
            var roles = await _roleDBService.FindRolesByTypeAsync(RoleShowTypeEnum.ISShow, "");
            var notShowRoles = await _roleDBService.FindRolesByTypeAsync(RoleShowTypeEnum.NotShow, "");
            roles.AddRange(notShowRoles);
            if (roles != null && roles.Count > 0)
            {
                foreach (var role in roles)
                {
                    ChangeRoleByLanguageConfig(role, request.Language);
                }
            }
            return roles;
        }

        private void ChangeRoleByLanguageConfig(RoleDTO role, string areaCode)
        {
            var langConfig = role.LanguageConfigs?.FirstOrDefault(x => x.Language == areaCode);
            if (langConfig != null)
            {
                role.RoleName = langConfig.RoleName;
                role.Description = langConfig.Description;
                role.FieldList = langConfig.FieldList;
            }
            else
            {
                role.RoleName = "";
                role.Description = "";
                role.FieldList = new List<string>();
            }
        }
    }
}
