using System.Linq;
using System.Threading.Tasks;
using WingInterfaceLibrary.Result.Upgrade;
using WingInterfaceLibrary.Request.Upgrade;
using WingServerCommon.Service;
using WingInterfaceLibrary.Interface;
using WingServerCommon.Interfaces.Cache;
using WingInterfaceLibrary.Request.Authentication;
using Newtonsoft.Json.Linq;
using System;
using WingServerCommon.Config;
using WingInterfaceLibrary.Enum;
using System.Collections.Generic;
using Newtonsoft.Json;
using WingUserModule.Entity;
using WingInterfaceLibrary.Notifications;
using WingInterfaceLibrary.Internal.Request;
using WingInterfaceLibrary.Enum.NotificationEnum;
using WingServerCommon.Log;
using WingInterfaceLibrary.DTO.Organization;
using WingInterfaceLibrary.Request.User;
using WingInterfaceLibrary.DTO.User;
using WingServerCommon.Mapper;
using JsonRpcLite.Services;

namespace WingUserModule.Service
{
    public partial class UserService : JsonRpcService, IUpgradeService
    {
        /// <summary>
        /// 获取检查更新配置
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<GetUpgradeInfoResult> GetUpgradeInfoAsync(GetUpgradeInfoRequest request)
        {
            var result = new GetUpgradeInfoResult()
            {
                UpgradeType = (UpgradeTypeEnum)Enum.Parse(typeof(UpgradeTypeEnum), _updateMode)
            };
            try
            {
                if (string.IsNullOrEmpty(request.CurrentVersion))
                {
                    ThrowCustomerException(CustomerRpcCode.WaitUpgradeIsEmpty, "Wait upgrade is empty");
                }
                var storageSettingsParam = ConfigurationManager.StorageSettings;
                if (storageSettingsParam == null)
                {
                    ThrowCustomerException(CustomerRpcCode.StorageSettingsEmpty, "Storage settings empty");
                }
                //对于异常版本号的特殊处理  2024-1-11  兼容ios异常版本号
                if (request.CurrentVersion == "3")
                {
                    request.CurrentVersion = "2.0.10.4484";
                }
                var sourceType = request.SourceType;
                var currentVersion = request.CurrentVersion;//传入版本
                var fileName = "Flyinsono_";
                var packageName = "Flyinsono_";
                var upgradeEnum = UpgradeEnum.Windows;//更新类型
                if (sourceType == UpgradeSourceTypeEnum.Client)
                {
                    if (request.Platform == UpgradePlatformEnum.PC)
                    {
                        upgradeEnum = UpgradeEnum.Windows;
                        var versions = request.CurrentVersion.Split(';');
                        //pc客户端传入FIS版本
                        var pcFISVersion = versions.FirstOrDefault(x => x.Contains("fis=")).Split("=")[1].Split("+")[0];
                        //pc客户端传入flyinsono版本
                        var pcflyinsonoVersion = versions.FirstOrDefault(x => x.Contains("Flyinsono=")).Split("=")[1].Split("+")[0];
                        var curFISVersion = new Version(pcFISVersion);
                        var newFISVersion = new Version(_pcFISVersion);
                        var curFlyinsoVersion = new Version(pcflyinsonoVersion);
                        var newFlyinsoVersion = new Version(_pcFlyinsonoVersion);
                        if (curFISVersion >= newFISVersion && curFlyinsoVersion >= newFlyinsoVersion)
                        {
                            result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                            return result;
                        }
                        else if (curFISVersion < newFISVersion)
                        {
                            var _allFisVersion = _pcFISVersion.Split('.').LastOrDefault();//2.0.0.547-->547
                            fileName += $"{upgradeEnum.ToString()}_{_allFisVersion}_{_pcFlyinsonoVersion}_All.zip";
                            result.UpgradeUpdateType = UpgradeUpdateTypeEnum.All;
                        }
                        else if (curFlyinsoVersion < newFlyinsoVersion)
                        {
                            fileName += upgradeEnum.ToString() + "_" + _pcFlyinsonoVersion + ".zip";
                            result.UpgradeUpdateType = UpgradeUpdateTypeEnum.Part;
                        }
                        packageName += upgradeEnum.ToString() + "_" + _pcFlyinsonoVersion + ".exe";
                        //PC Client最新版本返回的是flyinsono的版本号
                        result.NewVersion = _pcFlyinsonoVersion;
                    }
                    else if (request.Platform == UpgradePlatformEnum.Android)
                    {
                        upgradeEnum = UpgradeEnum.Android;
                        var curVersion = new Version(currentVersion);
                        var newVersion = new Version(_androidVersion);
                        if (curVersion < newVersion)
                        {
                            //需要更新
                            fileName += upgradeEnum.ToString() + "_" + _androidVersion + ".apk";
                            result.NewVersion = _androidVersion;
                        }
                        else
                        {
                            result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                            return result;
                        }
                    }
                    else if (request.Platform == UpgradePlatformEnum.IOS)
                    {
                        var curVersion = new Version(currentVersion);
                        var newVersion = new Version(_androidVersion);//IOS没有特定版本，因为IOS需要审核 android就是杏聆荟客户端最新版本
                        if (curVersion < newVersion)
                        {
                            //需要更新，只需返回最新版本
                            //result.UpgradeSourceUrl = "https://apps.apple.com/cn/app/${widget.iosAppId!}";
                            result.NewVersion = _androidVersion;
                        }
                        else
                        {
                            result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                            return result;
                        }
                    }
                    else
                    {
                        result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                        return result;
                    }
                }
                else if (sourceType == UpgradeSourceTypeEnum.SONOPOST)
                {
                    upgradeEnum = UpgradeEnum.SonoPost;

                    var curVersion = new Version(currentVersion);
                    var newVersion = new Version(_sonoPostVersion);
                    if (curVersion < newVersion)
                    {
                        //需要更新
                        fileName += upgradeEnum.ToString() + "_" + _sonoPostVersion + ".zip";
                        result.NewVersion = _sonoPostVersion;
                    }
                    else
                    {
                        result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                        return result;
                    }
                }
                else if (sourceType == UpgradeSourceTypeEnum.FISLib)
                {

                    upgradeEnum = UpgradeEnum.FISSDK;

                    var curVersion = new Version(currentVersion);
                    var newVersion = new Version(_windowsFISLibVersion);
                    if (curVersion < newVersion)
                    {
                        //需要更新
                        fileName += upgradeEnum.ToString() + "_" + _windowsFISLibVersion + ".zip";
                        result.NewVersion = _windowsFISLibVersion;
                    }
                    else
                    {
                        result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                        return result;
                    }
                }
                else
                {
                    result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
                    return result;
                }
                var serverType = (WingInterfaceLibrary.Enum.StorageServerEnum)Enum.Parse(typeof(WingInterfaceLibrary.Enum.StorageServerEnum), _storageServer.ToUpper());
                if (WingInterfaceLibrary.Enum.StorageServerEnum.VCS == serverType)
                {
                    result.UpgradeSourceUrl = storageSettingsParam.CFSServerUrl + "/SystemUpgradePackage/" + fileName;
                    //result.UpgradeCDNUrl = storageSettingsParam.CDNServerUrl + "/SystemUpgradePackage/" + fileName;
                    if (upgradeEnum == UpgradeEnum.Windows)
                    {
                        result.BackUpSourceUrl = storageSettingsParam.CFSServerUrl + "/SystemUpgradePackage/" + packageName;
                        //result.BackUpCDNUrl = storageSettingsParam.CDNServerUrl + "/SystemUpgradePackage/" + packageName;
                    }
                }
                else
                {
                    result.UpgradeSourceUrl = storageSettingsParam.CFSServerUrl + "/" + fileName;
                    result.UpgradeCDNUrl = storageSettingsParam.CDNServerUrl + "/" + fileName;
                    if (upgradeEnum == UpgradeEnum.Windows)
                    {
                        result.BackUpSourceUrl = storageSettingsParam.CFSServerUrl + "/" + packageName;
                        result.BackUpCDNUrl = storageSettingsParam.CDNServerUrl + "/" + packageName;
                    }
                }
                JObject json = JObject.Parse(_description);
                try
                {
                    var desc = json[request.Platform.ToString()]?[request.Language].ToString();
                    result.Description = desc;
                }
                catch (Exception ex)
                {
                    var desc = json[request.Platform.ToString()]?["en-US"].ToString();
                    result.Description = desc;
                }
            }
            catch (RpcException rpcEx)
            {
                throw rpcEx;
            }
            catch (Exception ex)
            {
                Logger.WriteLineError($"IUpgradeService GetUpgradeInfoAsync,CurrentVersion:{request.CurrentVersion},Language:{request.Language},Platform:{request.Platform},SourceType:{request.SourceType},error:{ex.ToString()}");
                result.UpgradeType = UpgradeTypeEnum.NoUpgrade;
            }
            return result;
        }

        /// <summary>
        /// 定时更新,发送更新通知，暂定5分钟
        /// </summary>
        private void WatchingUpgrade()
        {
            Task.Run(async () =>
            {
                while (true)
                {
                    await MonitorUpgradeHandler();
                    await Task.Delay(60 * 5 * 1000);
                }
            });
        }

        /// <summary>
        /// 启动后,发送更新通知
        /// </summary>
        private async Task MonitorUpgradeHandler()
        {
            var upgradeNotice = new UpgradeVersionNotification();
            //更新规则记录在配置文件
            var storageSettingsParam = ConfigurationManager.StorageSettings;
            if (storageSettingsParam == null)
            {
                ThrowCustomerException(CustomerRpcCode.StorageSettingsEmpty, "Storage settings empty");
            }
            //监控需更新的终端 发送通知  
            var userTokenList = CacheMaintenance.Instance.Get<ITokensManager>().Where(t => (t.Platform == (int)Platform.Windows || t.Platform == (int)Platform.Android) && t.AccountType == (int)AccountType.User && t.IsOnline)?.ToList();
            if (userTokenList?.Count > 0)
            {
                var upgradeType = (UpgradeTypeEnum)Enum.Parse(typeof(UpgradeTypeEnum), _updateMode);
                upgradeNotice.UpgradeType = upgradeType;
                var grayscaleOrganizations = new List<string>();
                var grayscalePersons = new List<string>();
                if (!string.IsNullOrEmpty(_grayscaleRules))
                {
                    JObject grayscaleJson = JObject.Parse(_grayscaleRules);
                    var grayscaleOrganizationsStr = grayscaleJson["Organizations"].ToString();
                    var grayscalePersonsStr = grayscaleJson["Persons"].ToString();
                    if (!string.IsNullOrEmpty(grayscalePersonsStr) && grayscalePersonsStr.Split(',').Length > 0)
                    {
                        grayscalePersons = grayscalePersonsStr.Split(',').ToList();
                    }
                    if (!string.IsNullOrEmpty(grayscaleOrganizationsStr) && grayscaleOrganizationsStr.Split(',').Length > 0)
                    {
                        grayscaleOrganizations = grayscaleOrganizationsStr.Split(',').ToList();
                    }
                }
                foreach (var userToken in userTokenList)
                {
                    try
                    {
                        if (string.IsNullOrWhiteSpace(userToken.InstallVersion))
                        {
                            continue;
                        }
                        var isPCNotified = false;
                        var isAndroidNotified = false;
                        //验证PC是否已经发送过消息
                        var findMessageRequest = new WingInterfaceLibrary.DB.Request.FindMessagesDBRequest()
                        {
                            ClientId = userToken.ClientId,
                            NotificationTypes = new List<NotificationTypeEnum>() { NotificationTypeEnum.VersionUpgradeNotification },
                            RelevanceCode = UpgradeEnum.Windows + "_" + _pcFlyinsonoVersion,
                            ReceiverType = ApplicantTypeEnum.Client,
                            TransactionType = TransactionTypeEnum.Upgrade
                        };
                        var findResult = await _messageDBService.GetMessageDBByConditionAsync(findMessageRequest);
                        if (findResult != null && !string.IsNullOrEmpty(findResult.Code))
                        {
                            isPCNotified = true;
                        }
                        //验证Android是否已经发送过消息
                        var findAndroidMessageRequest = new WingInterfaceLibrary.DB.Request.FindMessagesDBRequest()
                        {
                            ClientId = userToken.ClientId,
                            NotificationTypes = new List<NotificationTypeEnum>() { NotificationTypeEnum.VersionUpgradeNotification },
                            RelevanceCode = UpgradeEnum.Android + "_" + _androidVersion,
                            ReceiverType = ApplicantTypeEnum.Client,
                            TransactionType = TransactionTypeEnum.Upgrade
                        };
                        var findAndroidResult = await _messageDBService.GetMessageDBByConditionAsync(findAndroidMessageRequest);
                        if (findAndroidResult != null && !string.IsNullOrEmpty(findAndroidResult.Code))
                        {
                            isAndroidNotified = true;
                        }
                        if (isPCNotified && isAndroidNotified)
                        {
                            continue;//PC、安卓都已通知
                        }
                        var userInfo = await _userDBService.FindUserByCodeAsync(userToken.ClientId);
                        var organizationCode = userInfo.OrganizationCode;
                        if (grayscalePersons?.Count > 0 && grayscalePersons.Contains(userToken.ClientId))
                        {
                            //需要更新
                        }
                        else if (grayscaleOrganizations?.Count > 0 && grayscaleOrganizations.Contains(organizationCode))
                        {
                            //需要更新
                        }
                        else
                        {
                            if (!string.IsNullOrEmpty(_grayscaleRules))
                            {
                                //无需更新
                                continue;
                            }
                        }
                        var platform = (Platform)userToken.Platform;
                        var upgradeEnum = UpgradeEnum.Windows;//更新类型
                        var upgradePlatformEnum = UpgradePlatformEnum.PC;
                        var fileName = "Flyinsono_";
                        var packageName = "Flyinsono_";
                        var relevanceCode = "";
                        //比较版本
                        if (platform == Platform.Windows && !isPCNotified)
                        {
                            upgradeEnum = UpgradeEnum.Windows;
                            upgradePlatformEnum = UpgradePlatformEnum.PC;
                            var versions = userToken.InstallVersion.Split(';');
                            //pc客户端传入FIS版本
                            var pcFISVersion = versions.FirstOrDefault(x => x.Contains("fis=")).Split("=")[1].Split("+")[0];
                            //pc客户端传入flyinsono版本
                            var pcflyinsonoVersion = versions.FirstOrDefault(x => x.Contains("Flyinsono=")).Split("=")[1].Split("+")[0];
                            var curFISVersion = new Version(pcFISVersion);
                            var newFISVersion = new Version(_pcFISVersion);
                            var curFlyinsoVersion = new Version(pcflyinsonoVersion);
                            var newFlyinsoVersion = new Version(_pcFlyinsonoVersion);
                            if (curFISVersion >= newFISVersion && curFlyinsoVersion >= newFlyinsoVersion)
                            {
                                //无需更新
                                continue;
                            }
                            else if (curFISVersion < newFISVersion)
                            {
                                var _allFisVersion = _pcFISVersion.Split('.').LastOrDefault();//2.0.0.547-->547
                                fileName += $"{upgradeEnum.ToString()}_{_allFisVersion}_{_pcFlyinsonoVersion}_All.zip";
                                upgradeNotice.UpgradeUpdateType = UpgradeUpdateTypeEnum.All;
                            }
                            else if (curFlyinsoVersion < newFlyinsoVersion)
                            {
                                fileName += upgradeEnum.ToString() + "_" + _pcFlyinsonoVersion + ".zip";
                                upgradeNotice.UpgradeUpdateType = UpgradeUpdateTypeEnum.Part;
                            }
                            packageName += upgradeEnum.ToString() + "_" + _pcFlyinsonoVersion + ".exe";
                            //PC Client最新版本返回的是flyinsono的版本号
                            upgradeNotice.NewVersion = _pcFlyinsonoVersion;
                            relevanceCode = upgradeEnum.ToString() + "_" + _pcFlyinsonoVersion;
                        }
                        else if (platform == Platform.Android && !isAndroidNotified)
                        {
                            var curVersion = new Version(userToken.InstallVersion);
                            var androidConfigVersion = new Version(_androidVersion);
                            if (curVersion < androidConfigVersion)
                            {
                                //需要更新
                                upgradeEnum = UpgradeEnum.Android;
                                upgradePlatformEnum = UpgradePlatformEnum.Android;
                                //需要更新
                                fileName += upgradeEnum.ToString() + "_" + _androidVersion + ".apk";
                                upgradeNotice.NewVersion = _androidVersion;
                                relevanceCode = upgradeEnum.ToString() + "_" + _androidVersion;
                            }
                            else
                            {
                                continue;
                            }
                        }
                        else
                        {
                            continue;
                        }
                        var serverType = (WingInterfaceLibrary.Enum.StorageServerEnum)Enum.Parse(typeof(WingInterfaceLibrary.Enum.StorageServerEnum), _storageServer.ToUpper());
                        if (WingInterfaceLibrary.Enum.StorageServerEnum.VCS == serverType)
                        {
                            upgradeNotice.UpgradeSourceUrl = storageSettingsParam.CFSServerUrl + "/SystemUpgradePackage/" + fileName;
                            //upgradeNotice.UpgradeCDNUrl = storageSettingsParam.CDNServerUrl + "/SystemUpgradePackage/" + fileName;
                            if (upgradeEnum == UpgradeEnum.Windows)
                            {
                                upgradeNotice.BackUpSourceUrl = storageSettingsParam.CFSServerUrl + "/SystemUpgradePackage/" + packageName;
                                //upgradeNotice.BackUpCDNUrl = storageSettingsParam.CDNServerUrl + "/SystemUpgradePackage/" + packageName;
                            }
                        }
                        else
                        {
                            upgradeNotice.UpgradeSourceUrl = storageSettingsParam.CFSServerUrl + "/" + fileName;
                            upgradeNotice.UpgradeCDNUrl = storageSettingsParam.CDNServerUrl + "/" + fileName;
                            if (upgradeEnum == UpgradeEnum.Windows)
                            {
                                upgradeNotice.BackUpSourceUrl = storageSettingsParam.CFSServerUrl + "/SystemUpgradePackage/" + packageName;
                                upgradeNotice.BackUpCDNUrl = storageSettingsParam.CDNServerUrl + "/SystemUpgradePackage/" + packageName;
                            }
                        }

                        JObject json = JObject.Parse(_description);
                        var descList = new List<DescriptionInfoDTO>();
                        var languageConfigDatas = LanguageSetting.GetLanguageConfigParammeter("LanguageConfigs", true);
                        if (languageConfigDatas?.Count > 0)
                        {
                            foreach (var item in languageConfigDatas)
                            {
                                try
                                {
                                    descList.Add(new DescriptionInfoDTO()
                                    {
                                        Description = json[upgradePlatformEnum.ToString()][item].ToString(),
                                        LanguageCode = item
                                    });
                                }
                                catch(Exception ex)
                                {
                                    descList.Add(new DescriptionInfoDTO()
                                    {
                                        Description = json[upgradePlatformEnum.ToString()]["en-US"].ToString(),
                                        LanguageCode = item
                                    });
                                }
                            }
                        }
                        else
                        {
                            descList.Add(new DescriptionInfoDTO()
                            {
                                Description = json[upgradePlatformEnum.ToString()]["en-US"].ToString(),
                                LanguageCode = "en-US"
                            });
                        }
                        upgradeNotice.Descriptions = descList;
                        SendVersionUpgradeNotification(upgradeNotice, userToken.ClientId, ApplicantTypeEnum.Client, relevanceCode);
                    }
                    catch (Exception ex)
                    {
                        Logger.WriteLineWarn($"MonitorUpgradeHandler fail,userName:{userToken.AccountName},Platform:{userToken.Platform},InstallVersion:{userToken.InstallVersion},LoginSource:{userToken.LoginSource} ex:" + ex);
                    }
                }
            }
        }

        /// <summary>
        /// 发送通知
        /// </summary>
        /// <param name="upgradeNotice"></param>
        /// <param name="clientIds"></param>
        /// <param name="applicantType"></param>
        private async Task SendVersionUpgradeNotification(UpgradeVersionNotification upgradeNotice, string clientId, ApplicantTypeEnum applicantType, string relevanceCode)
        {
            //在线的通知用户   
            var notificationRequest = new SendNotificationRequest()
            {
                ClientId = clientId,
                Message = upgradeNotice,
                JsonMessage = Newtonsoft.Json.JsonConvert.SerializeObject(upgradeNotice),
                NotificationType = upgradeNotice.NotificationType,
                TransactionType = TransactionTypeEnum.Upgrade,
                RelevanceCode = relevanceCode,
                ReceiverType = applicantType
            };
            //发送消息
            var result = _notificationService.PostMessageAsync(notificationRequest);
        }

        /// <summary>
        /// 上报版本接口
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<bool> UpdateClientVersion(SetUpgradeInfoRequest request)
        {
            var result = false;
            var tokenInfo = CacheMaintenance.Instance.Get<ITokensManager>().Get(request.Token);
            if (tokenInfo != null)
            {
                if (tokenInfo.InstallVersion != request.CurrentVersion)
                {
                    result = await _authenticationService.SetTokenInstallVersionAsync(new SetTokenInstallVersionRequest
                    {
                        Token = request.Token,
                        InstallVersion = request.CurrentVersion,
                    });
                }
            }
            return result;
        }

        /// <summary>
        /// 灰度发布搜索组织
        /// </summary>
        /// <param name="request">请求对象</param>
        /// <returns>组织信息列表</returns>
        /// <show>false</shou>
        public async Task<List<OrganizationDTO>> SearchGrayscaleOrganizationsAsync(SearchOrganizationsRequest request)
        {
            var organizations = await _organizationDBService.SearchOrganizationsAsync(request.Keyword, request.ParentCode, "", (OrganizationTypeEnum)((int)request.OrganizationType));
            if (!string.IsNullOrWhiteSpace(request.ParentCode))
            {
                organizations = organizations.Where(f => f.ParentCode == request.ParentCode).ToList();
            }
            return organizations;
        }

        /// <summary>
        /// 灰度发布根据组织和岗位身份查询用户列表信息
        /// </summary>
        /// <param name="request">请求实体</param>
        /// <returns>用户信息</returns>
        /// <remarks>POST</remarks>
        /// <show>false</shou>
        public async Task<List<UserExtendDTO>> GetGrayscaleUserListAsync(GetUserListRequest request)
        {
            var dbRequest = new GetUserListDBRequest
            {
                Keyword = request.Keyword,
                OrganizationCode = request.OrganizationCode,
                OrganizationQueryType = request.OrganizationQueryType,
                RankCode = request.RankCode,
                PositionCode = request.PositionCode,
                ExceptUsers = new List<string>(),
                RoleCodes = request.RoleCodes
            };
            var users = await _userDBService.GetUserListAsync(dbRequest);
            var result = users.MappingTo<List<UserExtendDTO>>();
            var findAuthenticationRolesRequest = new FindAuthenticationRolesRequest { Language = request.Language };
            var roles = await FindAllRolesAsync(findAuthenticationRolesRequest);
            foreach (var item in result)
            {
                if (item.RoleCodes != null && item.RoleCodes.Count > 0)
                {
                    item.RoleName = roles.FirstOrDefault(x => x.RoleCode == item.RoleCodes[0])?.RoleName;
                }
            }
            return result;
        }

        /// <summary>
        /// 获取当前客户端配置版本号请求
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public async Task<string> GetFlyinsonoVersionAsync(GetFlyinsonoVersionRequest request)
        {
            var currentVersion = string.Empty;
            if (request.CurrentClientType == CurrentClientSourceTypeEnum.PC)
            {
                currentVersion = _pcFISVersion;
            }
            else if (request.CurrentClientType == CurrentClientSourceTypeEnum.Web)
            {
                currentVersion = _pcFlyinsonoVersion;
            }
            else if (request.CurrentClientType == CurrentClientSourceTypeEnum.SONOPOST)
            {
                currentVersion = _sonoPostVersion;
            }
            else if (request.CurrentClientType == CurrentClientSourceTypeEnum.FISLib)
            {
                currentVersion = _windowsFISLibVersion;
            }
            else if (request.CurrentClientType == CurrentClientSourceTypeEnum.Android)
            {
                currentVersion = _androidVersion;
            }
            return currentVersion;
        }
    }
}